#!/usr/bin/python

#=============================================================================================
# Test MLWHAM by performing statistical tests on a set of of 1D harmonic oscillators, for which
# the true free energy differences can be computed analytically.
#
# A number of replications of an experiment in which i.i.d. samples are drawn from a set of
# K harmonic oscillators are produced.  For each replicate, we estimate the dimensionless free
# energy differences and mean-square displacements (an observable), as well as their uncertainties.
#
# For a 1D harmonic oscillator, the potential is given by
#   V(x;K) = (K/2) * (x-x_0)**2
# where K denotes the spring constant.
#
# The equilibrium distribution is given analytically by
#   p(x;beta,K) = sqrt[(beta K) / (2 pi)] exp[-beta K (x-x_0)**2 / 2]
# The dimensionless free energy is therefore
#   f(beta,K) = - (1/2) * ln[ (2 pi) / (beta K) ]
#
#=============================================================================================
# Version control information
__version__ = "$Revision: 1.4 $"
#=============================================================================================
# Imports
import MLWHAM
from numpy import *
from numpy.random import *
#=============================================================================================

#=============================================================================================
# PARAMETERS
#=============================================================================================

K_k = array([1., 2., 3, 4., 1., 1.0]) # spring constants for each state
O_k = array([-1.0, 0.5, 1.0, -0.5, -1.0, -1.0]) # offsets for spring constants 
N_k = array([30, 40, 50, 0, 10, 0]) # number of samples from each state (can be zero for some states)
beta = 1.0 # inverse temperature for all simulations
nreplicates = 100 # number of replicates of experiment for testing uncertainty estimate
observe = 'position' # the observable, one of 'mean square displacement' or 'position'

#=============================================================================================
# MAIN
#=============================================================================================

# Determine number of simulations.
K = N_k.size
if K_k.shape != N_k.shape: raise "K_k and N_k must have same dimensions."

# Determine maximum number of samples to be drawn for any state.
N_max = max(N_k)

# Compute widths of sampled distributions.
# For a harmonic oscillator with spring constant K,
# x ~ Normal(x_0, sigma^2), where sigma = 1/sqrt(beta K)
sigma_k = (beta * K_k)**-0.5
print "Gaussian widths:"
print sigma_k

# Compute the absolute dimensionless free energies of each oscillator analytically.
# f = - ln( sqrt((2 pi)/(beta K)) )
print 'Computing dimensionless free energies analytically...'
f_k_analytical = - log( sqrt(2 * pi) * sigma_k )

# Compute true free energy differences.
Deltaf_ij_analytical = zeros([K,K], dtype = float64)
for i in range(0,K):
  for j in range(0,K):
    Deltaf_ij_analytical[i,j] = f_k_analytical[j] - f_k_analytical[i]

# Compute ensemble averages analytically 
if observe == 'mean square displacement':
  A_k_analytical = sigma_k**2          # mean square displacement
elif observe == 'position': 
  A_k_analytical  = O_k                 # observable is the position
else:
  raise "Observable %s not known." % observe

# DEBUG info
print "This script will perform %d replicates of an experiment where samples are drawn from %d harmonic oscillators." % (nreplicates, K)
print "The harmonic oscillators have equilibrium positions"
print O_k
print "and spring constants"
print K_k
print "and the following number of samples will be drawn from each (can be zero if no samples drawn):"
print N_k
print ""

# Conduct a number of replicates of the same experiment
replicates = [] # storage for one has for each replicate
for replicate_index in range(0,nreplicates):
  print "Performing replicate %d / %d" % (replicate_index+1, nreplicates)

  # Initialize a hash to store data for this replicate.
  replicate = { }
  
  #=============================================================================================
  # Generate independent data samples from K one-dimensional harmonic oscillators centered at q = 0.
  #=============================================================================================
  
  print 'generating samples...'

  # Initialize storage for samples.
  x_kn = zeros([K,N_max], dtype = float64) # x_kn[k,n] is the coordinate x of independent snapshot n of simulation k
  U_kln = zeros([K,K,N_max], dtype = float64) # U_kmt(k,m,t) is the value of the energy for snapshot t from simulation k at potential m

  # Generate samples.
  for k in range(0,K):
    # generate N_k[k] independent samples with spring constant K_k[k]
    x_kn[k,0:N_k[k]] = normal(O_k[k], sigma_k[k], [N_k[k]])
    
    # compute potential energy of all samples in all potentials
    for l in range(0,K):      
      U_kln[k,l,:] = (K_k[l]/2.0) * (x_kn[k,:]-O_k[l])**2
 
  #=============================================================================================
  # Estimate free energies and expectations.
  #=============================================================================================

  # Estimate free energies from simulation using MLWHAM.
  print "Estimating relative free energies from simulation (this may take a while)..."

  # Compute reduced potential energies.
  u_kln = beta * U_kln

  # Initialize the MLWHAM class, determining the free energies.
  #mlwham = MLWHAM.MLWHAM(u_kln, N_k, verbose = True, method = 'Newton-Raphson')  
  mlwham = MLWHAM.MLWHAM(u_kln, N_k)
  
  # Get matrix of dimensionless free energy differences and uncertainty estimate.
  print "Computing covariance matrix..."
  (Deltaf_ij_estimated, dDeltaf_ij_estimated) = mlwham.getFreeEnergyDifferences()
  
  # Compute error from analytical free energy differences.
  Deltaf_ij_error = Deltaf_ij_estimated - Deltaf_ij_analytical

  # DEBUG
  print "Deltaf_ij_analytical:"
  print Deltaf_ij_analytical
  print "Deltaf_ij_estimated:"
  print Deltaf_ij_estimated
  print "dDeltaf_ij_estimated:"
  print dDeltaf_ij_estimated
  print "Deltaf_ij_error:"
  print Deltaf_ij_error

  # Estimate the expectation of the mean-squared displacement at each condition.
  if observe == 'mean square displacement':
    A_kn = zeros([K,N_max], dtype = float64);
    for k in range(0,K):
      A_kn[k,0:N_k[k]] = (x_kn[k,0:N_k[k]] - O_k[k])**2 # observable is the squared displacement
  # observable for estimation is the position
  if observe == 'position': 
    A_kn = x_kn   

  (A_k_estimated, dA_k_estimated) = mlwham.computeExpectations(A_kn)

  # Compute error from analytical observable estimate.
  A_k_error = A_k_estimated - A_k_analytical

  # DEBUG
  print "A_k_estimated:"
  print A_k_estimated
  print "A_k_analytical:"
  print A_k_analytical
  print "dA_k_estimated:"
  print dA_k_estimated
  print "A_k_error:"
  print A_k_error
  print ""
  
  #=============================================================================================
  # Store data for this replicate.
  #=============================================================================================  
  replicate['Deltaf_ij_estimated'] = Deltaf_ij_estimated.copy()
  replicate['dDeltaf_ij_estimated'] = dDeltaf_ij_estimated.copy()
  replicate['Deltaf_ij_error'] = Deltaf_ij_error.copy()
  replicate['A_k_estimated'] = A_k_estimated.copy()
  replicate['dA_k_estimated'] = dA_k_estimated.copy()
  replicate['A_k_error'] = A_k_error.copy()
  replicates.append(replicate)

#=============================================================================================
# Store all data as a pickle for later analysis.
#=============================================================================================
import pickle
outfile = open('mlwham_test.pkl', 'wb')
pickle.dump(replicates, outfile)
outfile.close()

#=============================================================================================
# Analyze data.
# NOTE: This can be commented out if desired.
#=============================================================================================
#
# By Chebyshev's inequality, we should have
#   P(error >= alpha sigma) <= 1 / alpha^2
# so that a lower bound will be
#   P(error < alpha sigma) > 1 - 1 / alpha^2
# for any real multiplier 'k', where 'sigma' represents the computed uncertainty (as one standard deviation).
#
# If the error is normal, we should have
#   P(error < alpha sigma) = erf(alpha / sqrt(2))

print "The uncertainty estimates are tested in this section."
print "If the error is normally distributed, the actual error will be less than a multiplier 'alpha' times the computed uncertainty 'sigma' a fraction of time given by"
print "  P(error < alpha sigma) = erf(alpha / sqrt(2))"
print "For example, the true error should be less than 1.0 * sigma (one standard deviation) a total of 68% of the time, and less than 2.0 * sigma (two standard deviations) 95% of the time."
print "The observed fraction of the time that error < alpha sigma, and its uncertainty, is given as 'obs' and 'obs err' below."
print "This should be compared to the column labeled 'normal'."
print "A weak lower bound that holds regardless of how the error is distributed is given by Chebyshev's inequality, and is listed as 'cheby' below."
print "Uncertainty estimates are tested for both free energy differences and expectations."
print ""

# Free energies.
min_alpha = 0.1
max_alpha = 3.0
nalpha = 30
alpha_values = linspace(min_alpha, max_alpha, num = nalpha)
Pobs = zeros([nalpha], dtype = float64)
dPobs = zeros([nalpha], dtype = float64)
for alpha_index in range(0,nalpha):
  # Get alpha value.
  alpha = alpha_values[alpha_index]
  # Accumulate statistics across replicates
  a = 1.0
  b = 1.0
  nunique = K * (K-1) / 2  
  for replicate in replicates:    
    # Compute fraction of free energy differences where error <= alpha sigma
    for i in range(0,K):
      for j in range(0,i):
        if isnan(replicate['Deltaf_ij_error'][i,j]) or isnan(replicate['dDeltaf_ij_estimated'][i,j]):
          print "replicate %d"
          print "Deltaf_ij_error"
          print replicate['Deltaf_ij_error']
          print "dDeltaf_ij_estimated"
          print replicate['dDeltaf_ij_estimated']
          raise "isnan"
        else:                                                             
          b += 1.0 / nunique
          if abs(replicate['Deltaf_ij_error'][i,j]) <= alpha * replicate['dDeltaf_ij_estimated'][i,j]:
            a += 1.0 / nunique
  Pobs[alpha_index] = a / b
  dPobs[alpha_index] = sqrt( a*b / ((a+b)**2 * (a+b+1)) )

# Write error as a function of sigma.
from scipy.special import erf
print "Error vs. alpha for free energies"
print "%5s %10s %10s %10s %10s" % ('alpha', 'cheby', 'obs', 'obs err', 'normal')
for alpha_index in range(0,nalpha):
  alpha = alpha_values[alpha_index]
  print "%5.1f %10.6f %10.6f +- %10.6f %10.6f" % (alpha, 1. - 1./alpha**2, Pobs[alpha_index], dPobs[alpha_index], erf(alpha / sqrt(2.)))

# compute bias
bias = 0.0
rms_error = 0.0
denom = 0.0
for replicate in replicates:    
  for i in range(0,K):
    for j in range(0,i):
      bias += replicate['Deltaf_ij_error'][i,j]
      rms_error += replicate['Deltaf_ij_error'][i,j]**2      
      denom += 1.0
bias /= denom
rms_error = sqrt(rms_error / denom)
print "bias: %f / %f" % (bias, rms_error)

# Expectations.
min_alpha = 0.1
max_alpha = 3.0
nalpha = 30
alpha_values = linspace(min_alpha, max_alpha, num = nalpha)
Pobs = zeros([nalpha], dtype = float64)
dPobs = zeros([nalpha], dtype = float64)
for alpha_index in range(0,nalpha):
  # Get alpha value.
  alpha = alpha_values[alpha_index]
  # Accumulate statistics across replicates
  a = 1.0
  b = 1.0
  nunique = K
  for replicate in replicates:    
    # Compute fraction of expectations where error <= alpha sigma
    for i in range(0,K):
      b += 1.0 / nunique
      if abs(replicate['A_k_error'][i]) <= alpha * replicate['dA_k_estimated'][i]:
        a += 1.0 / nunique
  Pobs[alpha_index] = a / b
  dPobs[alpha_index] = sqrt( a*b / ((a+b)**2 * (a+b+1)) )
                
# Write error as a function of sigma.
print "Error vs. alpha for expectations"
print "%5s %10s %10s %10s %10s" % ('alpha', 'cheby', 'obs', 'obs err', 'normal')
for alpha_index in range(0,nalpha):
  alpha = alpha_values[alpha_index]
  print "%5.1f %10.6f %10.6f +- %10.6f %10.6f" % (alpha, 1. - 1./alpha**2, Pobs[alpha_index], dPobs[alpha_index], erf(alpha / sqrt(2.)))


